/*
 * Copyright (c) 2018 Red Hat, Inc.
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA
 * 02110-1301, USA. 
 */

#ifndef RELEASE_VERSIONS_H
#define RELEASE_VERSIONS_H

enum {
  OXYGEN_RELEASE_VERSION_NUMBER = 109583,
  FLUORINE_RELEASE_VERSION_NUMBER = 115838,
  NEON_RELEASE_VERSION_NUMBER = 120965,
  SODIUM_RELEASE_VERSION_NUMBER = 127441,
  MAGNESIUM_RELEASE_VERSION_NUMBER = 131337,
  HEAD_RELEASE_VERSION_NUMBER = 0,
  CURRENT_RELEASE_VERSION_NUMBER = MAGNESIUM_RELEASE_VERSION_NUMBER,
};

#endif /* not RELEASE_VERSIONS_H */
